
/*
	config stuff related to different stages of
	code optimization
*/

/* dont use __AX__ assignments to load values, but load them directly */
#define NO_AX_IN_MACROS

#define USEOPTIMIZED

/*
	some hardware related constants
*/

#ifdef __C64__
#define SCREENBUFFER1 0xe000
#define SCREENBUFFER2 0xe400
#define COLBUFFER 	  0xd800
#define VICBASE 0xd000

/* size of the screen/viewport */
#define SCREENX 40
#define SCREENY 25

/* use optimized assembler subroutines */
#define USEOPTIMIZEDASM
#define USEMIRRORING
//#define ENABLESPRITES

#endif

#ifdef __VIC20__
/* size of the screen/viewport */
#define SCREENX 23
#define SCREENY 21

/* use the standard conio lib for output */
#define USECONIO
#endif

#ifdef __PLUS4__
/* size of the screen/viewport */
#define SCREENX 40
#define SCREENY 25

/* use the standard conio lib for output */
#define USECONIO
#endif

/*
	some onstants related to the actual raycaster
*/

/* number of rays to cast, usually equals the screen width in pixels */
#define NUMRAYS   SCREENX

/* max steps to advance each ray. defines how far you can see */
#define MAXSTEPS  (16*8)

/* size of the world map */
#define WORLDX 16
#define WORLDY 16

#if 0

/*
	field of view in radiants

	note: dont make the FOV larger than about a quarter
	      of the circle (0x40 in radiants here) or things
		  will start looking strange at best.
*/
#define FIELDOFVIEW (0x100/4)
#define RAYSTEP     (FIELDOFVIEW/NUMRAYS)

#else
/*
	we want to achive that RAYSTEP is an integer,
	so instead we do this
*/
#define RAYSTEP 	1
#define FIELDOFVIEW	(NUMRAYS*RAYSTEP)

#endif

/*
	needed libs
*/

#include <conio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#ifdef __C64__
#include <c64.h>
#endif

/*
	prototypes
*/
void draw_mirror1(void);
void draw_mirror2(void);
void draw_colmirror(void);

#ifdef __C64__
void startirq(void);
void stopirq(void);

void irqincoffs(void);
void irqdecoffs(void);
#endif

#include "sinus.h"

#ifdef __C64__
#include "sprites.c"
#endif

/*************************************************************************************
 stuff for environment handling
*************************************************************************************/

/*
	the world map

	0x00 - empty cell
*/
const unsigned char world_blocks[WORLDY][WORLDX]=
{
{0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x04},
{0x02,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x04},
{0x02,0x00,0x01,0x04,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x01,0x01,0x00,0x04},
{0x02,0x00,0x02,0x03,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x01,0x00,0x04},
{0x02,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x00,0x01,0x00,0x04},
{0x02,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x00,0x04},
{0x02,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x80,0x00,0x00,0x01,0x00,0x04},
{0x02,0x00,0x00,0x00,0x01,0x02,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x00,0x04},
{0x02,0x00,0x00,0x00,0x03,0x04,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x00,0x04},
{0x02,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x81,0x00,0x01,0x00,0x04},
{0x02,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x00,0x04},
{0x02,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x00,0x04},
{0x02,0x00,0x04,0x03,0x00,0x00,0x00,0x00,0x00,0x00,0x82,0x00,0x01,0x01,0x00,0x04},
{0x02,0x00,0x01,0x02,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x01,0x01,0x00,0x04},
{0x02,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x04},
{0x02,0x03,0x03,0x03,0x03,0x03,0x03,0x03,0x03,0x03,0x03,0x03,0x03,0x03,0x03,0x04},
};
unsigned char worldptr_lo[WORLDY];
unsigned char worldptr_hi[WORLDY];

/*************************************************************************************
 stuff for camera handling
*************************************************************************************/

unsigned char  camera_tilt;
unsigned short camera_x;
unsigned short camera_y;

/*************************************************************************************
 stuff for drawing
*************************************************************************************/

unsigned char fisheye[SCREENX][MAXSTEPS];

//unsigned char *fisheyeptr[SCREENX];
unsigned char fisheyeptr_lo[SCREENX];
unsigned char fisheyeptr_hi[SCREENX];

#define SKYTEXTURE 		0x1f

unsigned char sky_project[MAXSTEPS]=
{
//0,1  ,2 ,3 ,4 ,5 ,6 ,7 , 8, 9,10,11,12,12,12,12,
0,0,0,0,0,0,0,0,
1,1,1,1,1,1,1,1,
2,2,2,2,2,2,2,2,
3,3,3,3,3,3,3,3,

4,4,4,4,4,4,4,4,
5,5,5,5,5,5,5,5,
6,6,6,6,6,6,6,6,
7,7,7,7,7,7,7,7,

8,8,8,8,8,8,8,8,
9,9,9,9,9,9,9,9,
10,10,10,10,10,10,10,10,
11,11,11,11,11,11,11,11,

12,12,12,12,12,12,12,12,
12,12,12,12,12,12,12,12,
12,12,12,12,12,12,12,12,
12,12,12,12,12,12,12,12,

//12,12,12,12,12,12,12,12,
//12,12,12,12,12,12,12,12,
//12,12,12,12,12,12,12,12,
//12,12,12,12,12,12,12,12,
};
unsigned char wall_project[MAXSTEPS];

unsigned char wall_project2[0x100]=
{
 0x00,0x00,0x00,0x01,0x01,0x01,0x02,0x02,
 0x02,0x03,0x03,0x03,0x04,0x04,0x04,0x05,
 0x05,0x05,0x06,0x06,0x06,0x07,0x07,0x07,
 0x08,0x08,0x08,0x09,0x09,0x09,0x0a,0x0a,

 0x0a,0x0b,0x0b,0x0b,0x0c,0x0c,0x0c,0x0d,
 0x0d,0x0d,0x0e,0x0e,0x0e,0x0f,0x0f,0x0f,
 0x10,0x10,0x10,0x11,0x11,0x11,0x12,0x12,
 0x12,0x12,0x12,0x13,0x13,0x13,0x14,0x14,

 0x14,0x14,0x15,0x15,0x16,0x16,0x17,0x17,
 0x18,0x18,0x19,0x19,0x1a,0x1a,0x1b,0x1b,
 0x1c,0x1c,0x1d,0x1d,0x1e,0x1e,0x1e,0x1e,

 0x1e,0x1e,0x1e,0x1e,0x1e,0x1e,0x1e,0x1e,
 0x1e,0x1e,0x1e,0x1e,0x1e,0x1e,0x1e,0x1e,
 0x1e,0x1e,0x1e,0x1e,0x1e,0x1e,0x1e,0x1e,
 0x1e,0x1e,0x1e,0x1e,0x1e,0x1e,0x1e,0x1e,
};

#ifdef USECONIO
unsigned char col_project[MAXSTEPS*8]=
{
    COLOR_WHITE,COLOR_WHITE,COLOR_WHITE,COLOR_WHITE,COLOR_WHITE,COLOR_WHITE,COLOR_WHITE,
    COLOR_WHITE,COLOR_WHITE,COLOR_WHITE,COLOR_WHITE,COLOR_WHITE,COLOR_WHITE,COLOR_WHITE,

    COLOR_YELLOW,COLOR_YELLOW,COLOR_YELLOW,COLOR_YELLOW,COLOR_YELLOW,COLOR_YELLOW,COLOR_YELLOW,
    COLOR_YELLOW,COLOR_YELLOW,COLOR_YELLOW,COLOR_YELLOW,COLOR_YELLOW,COLOR_YELLOW,COLOR_YELLOW,

    COLOR_CYAN,COLOR_CYAN,COLOR_CYAN,COLOR_CYAN,COLOR_CYAN,COLOR_CYAN,COLOR_CYAN,
    COLOR_CYAN,COLOR_CYAN,COLOR_CYAN,COLOR_CYAN,COLOR_CYAN,COLOR_CYAN,COLOR_CYAN,

    COLOR_GREEN,COLOR_GREEN,COLOR_GREEN,COLOR_GREEN,COLOR_GREEN,COLOR_GREEN,COLOR_GREEN,
    COLOR_GREEN,COLOR_GREEN,COLOR_GREEN,COLOR_GREEN,COLOR_GREEN,COLOR_GREEN,COLOR_GREEN,

    COLOR_RED,COLOR_RED,COLOR_RED,COLOR_RED,COLOR_RED,COLOR_RED,COLOR_RED,
    COLOR_RED,COLOR_RED,COLOR_RED,COLOR_RED,COLOR_RED,COLOR_RED,COLOR_RED,

    COLOR_BLUE,COLOR_BLUE,COLOR_BLUE,COLOR_BLUE,COLOR_BLUE,COLOR_BLUE,COLOR_BLUE,

    COLOR_BLUE,COLOR_BLUE,COLOR_BLUE,COLOR_BLUE,COLOR_BLUE,COLOR_BLUE,COLOR_BLUE,
    COLOR_BLUE,COLOR_BLUE,COLOR_BLUE,COLOR_BLUE,COLOR_BLUE,COLOR_BLUE,COLOR_BLUE,

    COLOR_BLUE,COLOR_BLUE,COLOR_BLUE,COLOR_BLUE,COLOR_BLUE,COLOR_BLUE,COLOR_BLUE,
    COLOR_BLUE,COLOR_BLUE,COLOR_BLUE,COLOR_BLUE,COLOR_BLUE,COLOR_BLUE,COLOR_BLUE,
};
#endif

#ifdef __C64__
unsigned char col_project[MAXSTEPS*8]=
{
 //0x01,0x0d,0x03,0x05,0x0c,0x08,0x0b,0x09,
 0x01+8,0x01+8,0x01+8,0x01+8,0x01+8,0x01+8,0x01+8,//0x01+8,
 0x01+8,0x01+8,0x01+8,0x01+8,0x01+8,0x01+8,0x01+8,//0x01+8,

 0x07+8,0x07+8,0x07+8,0x07+8,0x07+8,0x07+8,0x07+8,//0x07+8,
 0x07+8,0x07+8,0x07+8,0x07+8,0x07+8,0x07+8,0x07+8,//0x07+8,

 0x03+8,0x03+8,0x03+8,0x03+8,0x03+8,0x03+8,0x03+8,//0x03+8,
 0x03+8,0x03+8,0x03+8,0x03+8,0x03+8,0x03+8,0x03+8,//0x03+8,

 0x05+8,0x05+8,0x05+8,0x05+8,0x05+8,0x05+8,0x05+8,//0x05+8,
 0x05+8,0x05+8,0x05+8,0x05+8,0x05+8,0x05+8,0x05+8,//0x05+8,

 0x04+8,0x04+8,0x04+8,0x04+8,0x04+8,0x04+8,0x04+8,//0x04+8,
 0x04+8,0x04+8,0x04+8,0x04+8,0x04+8,0x04+8,0x04+8,//0x04+8,

 0x02+8,0x02+8,0x02+8,0x02+8,0x02+8,0x02+8,0x02+8,//0x02+8,
 0x02+8,0x02+8,0x02+8,0x02+8,0x02+8,0x02+8,0x02+8,//0x02+8,

 0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,

 0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,
 0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,

 0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,
 0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,

// 0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,
// 0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,

// 0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,
// 0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,0x06+8,
};
#endif

/*************************************************************************************
 stuff for the raycaster
*************************************************************************************/

// distance to wall(s) for each ray
unsigned char ray_d[NUMRAYS];



/*************************************************************************************
 world handling code
*************************************************************************************/

#if (WORLDX==16)&&(WORLDY==16)
/* table for shifting 4 bits */
unsigned char shifttab[0x10]=
{
    0x00,0x10,0x20,0x30,0x40,0x50,0x60,0x70,
    0x80,0x90,0xa0,0xb0,0xc0,0xd0,0xe0,0xf0
};
#endif

void world_setup(void)
{
#if (WORLDX!=16)||(WORLDY!=16)
unsigned char i,ii;
	for(i=0;i<WORLDY;++i)
	{
		worldptr_lo[i]=((unsigned char)(&world_blocks[i][0]));
		worldptr_hi[i]=((unsigned char)(((unsigned short)&world_blocks[i][0])>>8));
	}
#endif
}

/*
	getblock(unsigned short x,unsigned short y)

	returns a block from the world map
*/

#if 0

unsigned char getblock(unsigned short x,unsigned short y)
{
	return world_blocks[y>>SINSHIFT][x>>SINSHIFT];
}

#endif

#ifdef NO_AX_IN_MACROS

#if (WORLDX==16)&&(WORLDY==16)
/* special case version for 16*16 map */
#define getblock(_x,_y) (unsigned char) \
	( \
	__asm__ ("lda %v+1", _x ), \
	__asm__ ("ldx %v+1", _y ), \
	__asm__ ("ora %v,x", shifttab ), \
	__asm__ ("tax"), \
	__asm__ ("lda %v,x", world_blocks), \
	__asm__ ("ldx #%b", 0), \
     __AX__)
#else
/* generic version */
#define getblock(_x,_y) (unsigned char) \
	( \
	__asm__ ("ldy %v+1", _x ), \
	__asm__ ("ldx %v+1", _y ), \
	__asm__ ("lda %v,x", worldptr_lo ), \
	__asm__ ("sta ptr1"), \
	__asm__ ("lda %v,x", worldptr_hi ), \
	__asm__ ("sta ptr1+1"), \
	__asm__ ("lda (ptr1),y"), \
	__asm__ ("ldx #%b", 0), \
     __AX__)
#endif

#else

/* generic version */
#define getblock(_x,_y) (unsigned char) \
	( \
	__AX__= (_y), \
	__asm__ ("txa"), \
	__asm__ ("tay"), \
	__AX__= (_x), \
	__asm__ ("lda %v,x", worldptr_lo ), \
	__asm__ ("sta ptr1"), \
	__asm__ ("lda %v,x", worldptr_hi ), \
	__asm__ ("sta ptr1+1"), \
	__asm__ ("lda (ptr1),y"), \
	__asm__ ("ldx #%b", 0), \
     __AX__)

#endif // NO_AX_IN_MACROS

#ifdef NO_AX_IN_MACROS

#define getfisheye(_x,_d) (unsigned char) \
	(\
	__asm__ ("ldy %v", _x), \
	__asm__ ("lda %v,y", fisheyeptr_lo ), \
	__asm__ ("sta ptr1"), \
	__asm__ ("lda %v,y", fisheyeptr_hi ), \
	__asm__ ("sta ptr1+1"), \
	__asm__ ("ldy %v", _d), \
	__asm__ ("lda (ptr1),y"), \
     __AX__)

#else

#define getfisheye(_x,_d) (unsigned char) \
	(__AX__= (_d), \
	__asm__ ("tay"), \
	__AX__= (_x), \
	__asm__ ("tax"), \
	__asm__ ("lda %v,x", fisheyeptr_lo ), \
	__asm__ ("sta ptr1"), \
	__asm__ ("lda %v,x", fisheyeptr_hi ), \
	__asm__ ("sta ptr1+1"), \
	__asm__ ("lda (ptr1),y"), \
     __AX__)

#endif

#if 0
#define getfisheye(_x,_d) fisheye[_x][_d];
#define getfisheye(_x,_d) *(fisheyeptr[_x]+(_d));
#endif


/*************************************************************************************
 raycaster code
*************************************************************************************/

#ifdef NO_AX_IN_MACROS

#define uaddsc(_a,_b) \
	__asm__ ("lda %v", _a); \
	__asm__ ("clc"); \
	__asm__ ("adc %v", _b); \
	__asm__ ("sta %v", _a); \
	__asm__ ("bcc @L%s", __LINE__); \
	__asm__ ("inc %v+1", _a); \
	__asm__ ("@L%s:", __LINE__);


#define usubsc(_a,_b) \
	__asm__ ("lda %v", _a); \
	__asm__ ("sec"); \
	__asm__ ("sbc %v", _b); \
	__asm__ ("sta %v", _a); \
	__asm__ ("bcs @L%s", __LINE__); \
	__asm__ ("dec %v+1", _a); \
	__asm__ ("@L%s:", __LINE__);

#define ulsrc(_a) \
	__asm__ ("lsr %v", _a);

#else

#endif // NO_AX_IN_MACROS

#if 0

#define uaddsc(_a,_b) _a+=(_b)
#define usubsc(_a,_b) _a-=(_b)

#endif

void cast_rays(void)
{
register unsigned short x,y;
register unsigned char xx,yy;
register unsigned char steps;
register unsigned char cx;
register unsigned char i;
register unsigned char temp;

	cx=camera_tilt-(FIELDOFVIEW/2);
	for(i=0;i<NUMRAYS;++i)
	{
		x=camera_x;
		y=camera_y;
		xx=SINUS(cx);
		yy=COSINUS(cx);
		if(cx<0x40)
		{
			for(steps=8;steps<((MAXSTEPS-(1*8)));steps+=8)
			{
				/* advance ray */
				uaddsc(x,xx);
				uaddsc(y,yy);

#ifdef ENABLESPRITES
				if(((temp=getblock(x,y))&0x80)==0x80)
				{
					spriteset(temp,i,steps);
				}
				/* check if we hit a solid block */
				else if(temp!=0)
#else
				if(getblock(x,y)!=0)
#endif
				
				{
					// we know the block, now find out the fraction
					// go 4back
					ulsrc(xx);usubsc(x,xx);
					ulsrc(yy);usubsc(y,yy);

					if(getblock(x,y)!=0)
					{
						// go 2back
						ulsrc(xx);usubsc(x,xx);
						ulsrc(yy);usubsc(y,yy);
						// we are 6back
						if(getblock(x,y)!=0)
						{
							// go 1back
							ulsrc(xx);usubsc(x,xx);
							ulsrc(yy);usubsc(y,yy);
							// we are 7back
							if(getblock(x,y)!=0)
							{
								steps-=8;
							}
							else
							{
								steps-=7;
							}
						}
						else
						{
							// go 1forth
							ulsrc(xx);uaddsc(x,xx);
							ulsrc(yy);uaddsc(y,yy);
							// we are 5back
							if(getblock(x,y)!=0)
							{
								steps-=6;
							}
							else
							{
								steps-=5;
							}
						}
					}
					else
					{
						// go 2forth
						ulsrc(xx);uaddsc(x,xx);
						ulsrc(yy);uaddsc(y,yy);
						// we are 2back
						if(getblock(x,y)!=0)
						{
							// go 1back
							ulsrc(xx);usubsc(x,xx);
							ulsrc(yy);usubsc(y,yy);
							// we are 3back
							if(getblock(x,y)!=0)
							{
								steps-=4;
							}
							else
							{
								steps-=3;
							}
						}
						else
						{
							// go 1forth
							ulsrc(xx);uaddsc(x,xx);
							ulsrc(yy);uaddsc(y,yy);
							// we are 1back
							if(getblock(x,y)!=0)
							{
								steps-=2;
							}
							else
							{
								//steps-=1;
								--steps;
							}
						}
					}
					break;
				}
			}
		}
		else if(cx<0x80)
		{
			for(steps=8;steps<((MAXSTEPS-(1*8)));steps+=8)
			{
				/* advance ray */
				uaddsc(x,xx);
				usubsc(y,yy);

				/* check if we hit a solid block */
				if(getblock(x,y)!=0)
				{
					// we know the block, now find out the fraction

					// go 4back
					ulsrc(xx);usubsc(x,xx);
					ulsrc(yy);uaddsc(y,yy);

					if(getblock(x,y)!=0)
					{
						// go 2back
						ulsrc(xx);usubsc(x,xx);
						ulsrc(yy);uaddsc(y,yy);
						// we are 6back
						if(getblock(x,y)!=0)
						{
							// go 1back
							ulsrc(xx);usubsc(x,xx);
							ulsrc(yy);uaddsc(y,yy);
							// we are 7back
							if(getblock(x,y)!=0)
							{
								steps-=8;
							}
							else
							{
								steps-=7;
							}
						}
						else
						{
							// go 1forth
							ulsrc(xx);uaddsc(x,xx);
							ulsrc(yy);usubsc(y,yy);
							// we are 5back
							if(getblock(x,y)!=0)
							{
								steps-=6;
							}
							else
							{
								steps-=5;
							}
						}
					}
					else
					{
						// go 2forth
						ulsrc(xx);uaddsc(x,xx);
						ulsrc(yy);usubsc(y,yy);
						// we are 2back
						if(getblock(x,y)!=0)
						{
							// go 1back
							ulsrc(xx);usubsc(x,xx);
							ulsrc(yy);uaddsc(y,yy);
							// we are 3back
							if(getblock(x,y)!=0)
							{
								steps-=4;
							}
							else
							{
								steps-=3;
							}
						}
						else
						{
							// go 1forth
							ulsrc(xx);uaddsc(x,xx);
							ulsrc(yy);usubsc(y,yy);
							// we are 1back
							if(getblock(x,y)!=0)
							{
								steps-=2;
							}
							else
							{
								steps-=1;
							}
						}
					}
					break;
				}
			}
		}
		else if(cx<0xc0)
		{
			for(steps=8;steps<((MAXSTEPS-(1*8)));steps+=8)
			{
				/* advance ray */
				usubsc(x,xx);
				usubsc(y,yy);

				/* check if we hit a solid block */
				if(getblock(x,y)!=0)
				{
					// we know the block, now find out the fraction
					// go 4back
					ulsrc(xx);uaddsc(x,xx);
					ulsrc(yy);uaddsc(y,yy);
					if(getblock(x,y)!=0)
					{
						// go 2back
						ulsrc(xx);uaddsc(x,xx);
						ulsrc(yy);uaddsc(y,yy);
						// we are 6back
						if(getblock(x,y)!=0)
						{
							// go 1back
							ulsrc(xx);uaddsc(x,xx);
							ulsrc(yy);uaddsc(y,yy);
							// we are 7back
							if(getblock(x,y)!=0)
							{
								steps-=8;
							}
							else
							{
								steps-=7;
							}
						}
						else
						{
							// go 1forth
							ulsrc(xx);usubsc(x,xx);
							ulsrc(yy);usubsc(y,yy);
							// we are 5back
							if(getblock(x,y)!=0)
							{
								steps-=6;
							}
							else
							{
								steps-=5;
							}
						}
					}
					else
					{
						// go 2forth
						ulsrc(xx);usubsc(x,xx);
						ulsrc(yy);usubsc(y,yy);
						// we are 2back
						if(getblock(x,y)!=0)
						{
							// go 1back
							ulsrc(xx);uaddsc(x,xx);
							ulsrc(yy);uaddsc(y,yy);
							// we are 3back
							if(getblock(x,y)!=0)
							{
								steps-=4;
							}
							else
							{
								steps-=3;
							}
						}
						else
						{
							// go 1forth
							ulsrc(xx);usubsc(x,xx);
							ulsrc(yy);usubsc(y,yy);
							// we are 1back
							if(getblock(x,y)!=0)
							{
								steps-=2;
							}
							else
							{
								steps-=1;
							}
						}
					}
					break;
				}
			}
		}
		else
		{
			for(steps=8;steps<((MAXSTEPS-(1*8)));steps+=8)
			{
				/* advance ray */
				usubsc(x,xx);
				uaddsc(y,yy);

				/* check if we hit a solid block */
				if(getblock(x,y)!=0)
				{
					// we know the block, now find out the fraction

					// go 4back
					ulsrc(xx);uaddsc(x,xx);
					ulsrc(yy);usubsc(y,yy);
					if(getblock(x,y)!=0)
					{
						// go 2back
						ulsrc(xx);uaddsc(x,xx);
						ulsrc(yy);usubsc(y,yy);
						// we are 6back
						if(getblock(x,y)!=0)
						{
							// go 1back
							ulsrc(xx);uaddsc(x,xx);
							ulsrc(yy);usubsc(y,yy);
							// we are 7back
							if(getblock(x,y)!=0)
							{
								steps-=8;
							}
							else
							{
								steps-=7;
							}
						}
						else
						{
							// go 1forth
							ulsrc(xx);usubsc(x,xx);
							ulsrc(yy);uaddsc(y,yy);
							// we are 5back
							if(getblock(x,y)!=0)
							{
								steps-=6;
							}
							else
							{
								steps-=5;
							}
						}
					}
					else
					{
						// go 2forth
						ulsrc(xx);usubsc(x,xx);
						ulsrc(yy);uaddsc(y,yy);
						// we are 2back
						if(getblock(x,y)!=0)
						{
							// go 1back
							ulsrc(xx);uaddsc(x,xx);
							ulsrc(yy);usubsc(y,yy);
							// we are 3back
							if(getblock(x,y)!=0)
							{
								steps-=4;
							}
							else
							{
								steps-=3;
							}
						}
						else
						{
							// go 1forth
							ulsrc(xx);usubsc(x,xx);
							ulsrc(yy);uaddsc(y,yy);
							// we are 1back
							if(getblock(x,y)!=0)
							{
								steps-=2;
							}
							else
							{
								steps-=1;
							}
						}
					}
					break;
				}
			}
		}

		// remember number of steps (==distance)
		ray_d[i]=steps;
		// rotate to next ray
		cx+=RAYSTEP;
	}
}

/*************************************************************************************
 drawing code
*************************************************************************************/

#ifdef NO_AX_IN_MACROS

#define getbuffer(_b,_i,_buf) \
	( \
	__asm__ ("lda %v", _i), \
	__asm__ ("sta %v", _b), \
	__asm__ ("lda #>%w", _buf), \
	__asm__ ("sta %v+1", _b) \
     )
#else

#define getbuffer(_b,_i,_buf) \
	(__AX__ = (_i), \
	__asm__ ("ldx #>%w", _buf), \
	(_b) = (unsigned char*) __AX__ \
     )

#endif

/*
	setcolor(x,c)

	write a color to color-buffer
*/
#ifdef NO_AX_IN_MACROS

#define setcolor(_i,_c) \
	(\
	__asm__ ("ldx %v", _c), \
	__asm__ ("lda %v,x", col_project), \
	__asm__ ("ldx %v", _i), \
	__asm__ ("sta %w,x", COLBUFFER) \
     )

#else

#define setcolor(_i,_c) \
	(__AX__ = (_i), \
	__asm__ ("tay"), \
	 __AX__ = (_c), \
	__asm__ ("tax"), \
	__asm__ ("lda %v,x", col_project), \
	__asm__ ("sta %w,y", COLBUFFER) \
     )

#endif

//((unsigned char*)(COLBUFFER))[i]=col_project[temp2];

void project_setup(void)
{
unsigned char i;
	for(i=0;i<(MAXSTEPS);++i)
	{
		wall_project[i]=wall_project2[i]|(i<<5);
	}
}



void fisheye_setup(void)
{
unsigned char i,ii;
unsigned char cx;
signed short temp;

	cx=-(FIELDOFVIEW/2);

	//cx=0;

	for(i=0;i<SCREENX;++i)
	{
//		printf("\n\r%d:",i);
//		fisheyeptr[i]=&fisheye[i][0];
		fisheyeptr_lo[i]=((unsigned char)(&fisheye[i][0]));
		fisheyeptr_hi[i]=((unsigned char)(((unsigned short)&fisheye[i][0])>>8));
		for(ii=0;ii<MAXSTEPS;++ii)
		{
			temp=(MAXSTEPS-1)-((((signed short)MAXSTEPS-ii)<<SINSHIFT)/(signed short)COSINUS2(cx));

			/* clamp values since the above might not be very accurate */
			if(temp<0) temp = 0;
			if(temp>MAXSTEPS-1) temp = MAXSTEPS-1;

			fisheye[i][ii]=temp;

			//fisheye[i][ii]=sky_project[fisheye[i][ii]];

//			fisheye[i][ii]=(((signed short)MAXSTEPS-ii)<<8)/COSINUS(cx);
//			fisheye[i][ii]=MAXSTEPS-((((signed short)ii)<<8)/COSINUS(cx));
//			fisheye[i][ii]=(((signed short)ii)<<8)/COSINUS(cx);
//			fisheye[i][ii]=COSINUS(cx)/(ii<<8);
		//	if(ii==(MAXSTEPS-1))
	//		printf ("%x:%x:%x:%x \n",cx,((signed short)ii)<<8,COSINUS(cx),fisheye[i][ii]);
		}
//while(!kbhit()){};cgetc();
		cx+=RAYSTEP;
	}

	//exit(0);
}

void __fastcall__ draw_line(unsigned char i,unsigned char n,unsigned char d);
void __fastcall__ draw_line1(unsigned char i,unsigned char n,unsigned char d);

#ifdef USECONIO
void __fastcall__ draw_line(unsigned char i,unsigned char n,unsigned char d)
{
unsigned char temp;

    gotoxy(i,0);
    textcolor(COLOR_BLACK);
    temp=sky_project[d];
    cvline(temp);
    textcolor(col_project[d]);
    cvline(SCREENY-(temp<<1));
    textcolor(COLOR_BLACK);
    cvline(temp);
}
void __fastcall__ draw_line1(unsigned char i,unsigned char n,unsigned char d)
{
unsigned char temp;

    gotoxy(i,0);
    textcolor(COLOR_BLACK);
    temp=sky_project[d];
    cvline(temp);
    textcolor(col_project[d]);
    cvline(SCREENY-(temp<<1));
    textcolor(COLOR_BLACK);
    cvline(temp);
}
#endif

void draw_lines1(void)
{
register unsigned char *b;
register unsigned char temp3;
register unsigned char temp1;
register unsigned char y;
register unsigned char i;

	for(i=0;i<SCREENX;++i)
	{

#if 0
		temp1=temp3=ray_d[i]; 		   /* line distance */
		//setcolor(i,temp3);
		temp3=getfisheye(i,temp3); /* fisheye correction */

		//getbuffer(b,i,SCREENBUFFER1);


//		y=temp1=sky_project[temp3];
		//y=sky_project[temp3];
		//for(;temp1>0;--temp1)
		//{
		//	*b=SKYTEXTURE;
		//	b+=SCREENX;
		//}
		//*b=wall_project[temp3];

		//temp3=wall_project2[temp3];



		//b+=SCREENX;
		//for(;y<(SCREENY/2);++y)
		//{
		//	*b=temp3;
		//	b+=SCREENX;
		//}
		

#endif

 
 		temp1=ray_d[i]; 		   /* line distance */
		temp3=getfisheye(i,temp1); /* fisheye correction */
		draw_line(i,temp1,temp3);
	}
}

void draw_lines2(void)
{
register unsigned char *b;
register unsigned char temp3;
register unsigned char temp1;
register unsigned char y;
register unsigned char i;

	for(i=0;i<SCREENX;++i)
	{
#if 0
		temp3=ray_d[i]; 		   /* line distance */
		setcolor(i,temp3);
		temp3=getfisheye(i,temp3); /* fisheye correction */

		getbuffer(b,i,SCREENBUFFER2);

		y=temp1=sky_project[temp3];
		for(;temp1>0;--temp1)
		{
			*b=SKYTEXTURE;
			b+=SCREENX;
		}
		*b=wall_project[temp3];
		temp3=wall_project2[temp3];
		b+=SCREENX;
		for(;y<(SCREENY/2);++y)
		{
			*b=temp3;
			b+=SCREENX;
		}
#endif

		temp1=ray_d[i]; 		   /* line distance */
		temp3=getfisheye(i,temp1); /* fisheye correction */
		draw_line1(i,temp1,temp3);


	}
}


#ifndef USEOPTIMIZEDASM
void draw_mirror1(void)
{
#ifdef USEMIRRORING
unsigned short i,ii;
	for(i=0,ii=((SCREENY-1)*40);i<((SCREENY/2)*40);i+=40,ii-=40)
	{
		memcpy(&fb[ii],((unsigned char*)SCREENBUFFER1)[i],40);
		memcpy(&cb[ii],((unsigned char*)COLBUFFER)[i],40);
	}
#endif
}
void draw_mirror2(void)
{
#ifdef USEMIRRORING
unsigned short i,ii;
	for(i=0,ii=((SCREENY-1)*40);i<((SCREENY/2)*40);i+=40,ii-=40)
	{
		memcpy(&fb[ii],((unsigned char*)SCREENBUFFER2)[i],40);
		memcpy(&cb[ii],((unsigned char*)COLBUFFER)[i],40);
	}
#endif
}
#endif

unsigned char mychars[32*8]=
{
// 0x00
	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff, // full colram
	0xff,0xff,0xff,0xfe,0xff,0xff,0xff,0xfe,
	0xff,0xef,0xff,0xfe,0xff,0xef,0xff,0xfe,
	0xbf,0xef,0xff,0xfe,0xbf,0xef,0xff,0xfe,
// 0x04
	0xbf,0xef,0xfb,0xfe,0xbf,0xef,0xfb,0xfe,
	0xbf,0xef,0xfb,0xee,0xbf,0xef,0xfb,0xee,
	0xbf,0xee,0xfb,0xee,0xbf,0xee,0xfb,0xee,
	0xbb,0xee,0xbb,0xee,0xbb,0xee,0xbb,0xee, // colram/muco2
// 0x08
	0x7b,0xee,0xbb,0xee,0xbb,0xee,0xbb,0xde,
	0x7b,0xee,0xbb,0xee,0xbb,0xee,0xbb,0xde,
	0x7b,0xde,0xbb,0xed,0x7b,0xde,0xbb,0xde,
	0x7b,0xde,0xb7,0xed,0x7b,0xde,0xb7,0xde,
// 0x0c
	0x7b,0xde,0x77,0xed,0x77,0xde,0x77,0xde,
	0x7b,0xdd,0x77,0xed,0x77,0xdd,0x77,0xde,
	0x77,0xdd,0x77,0xdd,0x77,0xdd,0x77,0xdd, // colram/muco1
	0xd5,0x77,0x5d,0x77,0xd5,0x77,0x5d,0x77,
// 0x10
	0xd5,0x77,0x5d,0x77,0xd5,0x77,0x5d,0x77,
	0xd5,0x75,0x5d,0x57,0xd5,0x75,0x5d,0x57,
	0xd5,0x75,0x5d,0x57,0xd5,0x75,0x5d,0x57,
	0x55,0x75,0x55,0x55,0x55,0x75,0x55,0x57,
// 0x14
	0x55,0x75,0x55,0x55,0x55,0x75,0x55,0x57,
	0x55,0x55,0x55,0x55,0x55,0x55,0x55,0x55,  // full muco1
	0x55,0x51,0x55,0x51,0x55,0x51,0x55,0x51,
	0x54,0x51,0x45,0x51,0x54,0x51,0x45,0x51,
// 0x18
	0x54,0x51,0x45,0x51,0x54,0x51,0x45,0x51,
	0x54,0x51,0x45,0x51,0x54,0x51,0x45,0x51,
	0x44,0x11,0x44,0x11,0x44,0x11,0x44,0x11,   // muco1/bg
	0x40,0x11,0x04,0x11,0x40,0x11,0x04,0x11,
// 0x1c
	0x40,0x10,0x04,0x01,0x40,0x10,0x04,0x01,
	0x40,0x10,0x04,0x01,0x40,0x10,0x04,0x01,
	0x00,0x10,0x00,0x00,0x00,0x10,0x00,0x01,

//	0x75,0x17,0x75,0x17,0x75,0x17,0x75,0x17, // full colram
	0x8a,0xaa,0xaa,0xaa,0xab,0xaa,0xaa,0xaa,  // full muco1
//	0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff, // full colram
//	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,   // full bg
};

unsigned char oldchr;

void screen_setup(void)
{
#ifdef __C64__
int i,ii;
unsigned char c;
	oldchr=*(unsigned char*)(VICBASE+0x18);
	*(unsigned char*)0xdd00=0;
	for(i=0;i<(32*8);++i)
	{
     		c=mychars[i];
			*(unsigned char*)(0xf000+i)=c;
			if((i&0x07)==0) c=mychars[(0x1f*8)+0];
			*(unsigned char*)(0xf100+i)=c;
			if((i&0x07)==1) c=mychars[(0x1f*8)+1];
			*(unsigned char*)(0xf200+i)=c;
			if((i&0x07)==2) c=mychars[(0x1f*8)+2];
			*(unsigned char*)(0xf300+i)=c;
			if((i&0x07)==3) c=mychars[(0x1f*8)+3];
			*(unsigned char*)(0xf400+i)=c;
			if((i&0x07)==4) c=mychars[(0x1f*8)+4];
			*(unsigned char*)(0xf500+i)=c;
			if((i&0x07)==5) c=mychars[(0x1f*8)+5];
			*(unsigned char*)(0xf600+i)=c;
			if((i&0x07)==6) c=mychars[(0x1f*8)+6];
			*(unsigned char*)(0xf700+i)=c;

     		c=mychars[i];
			*(unsigned char*)(0xe800+i)=c;
			if((i&0x07)==7) c=mychars[(0x1f*8)+0];
			*(unsigned char*)(0xe900+i)=c;
			if((i&0x07)==6) c=mychars[(0x1f*8)+1];
			*(unsigned char*)(0xea00+i)=c;
			if((i&0x07)==5) c=mychars[(0x1f*8)+2];
			*(unsigned char*)(0xeb00+i)=c;
			if((i&0x07)==4) c=mychars[(0x1f*8)+3];
			*(unsigned char*)(0xec00+i)=c;
			if((i&0x07)==3) c=mychars[(0x1f*8)+4];
			*(unsigned char*)(0xed00+i)=c;
			if((i&0x07)==2) c=mychars[(0x1f*8)+5];
			*(unsigned char*)(0xee00+i)=c;
			if((i&0x07)==1) c=mychars[(0x1f*8)+6];
			*(unsigned char*)(0xef00+i)=c;
}

	/* clear the framebuffers */
	for(i=0;i<1000;++i)
	{
		*(unsigned char*)(SCREENBUFFER1+i)=0;
		*(unsigned char*)(SCREENBUFFER2+i)=0;
		*(unsigned char*)(COLBUFFER+i)=0;
	}
	*(unsigned char*)(VICBASE+0x16)=0xd8;
	*(unsigned char*)(VICBASE+0x22)=0x0b;
	*(unsigned char*)(VICBASE+0x23)=0x0c;
	startirq();
#endif

#ifdef USECONIO
    revers(1);
#endif

}

void screen_reset(void)
{
#ifdef __C64__
	stopirq();
	*(unsigned char*)(VICBASE+0x18)=oldchr;
	*(unsigned char*)0xdd00=3;
	*(unsigned char*)(VICBASE+0x16)=0xc8;
#endif
}

void screen_setfb1(void)
{
#ifdef __C64__
	*(unsigned char*)(VICBASE+0x18)=((*(unsigned char*)(VICBASE+0x18))&0x0f)|0x80;
#endif
}

void screen_setfb2(void)
{
#ifdef __C64__
	*(unsigned char*)(VICBASE+0x18)=((*(unsigned char*)(VICBASE+0x18))&0x0f)|0x90;
#endif
}

/*************************************************************************************
 camera handling code
*************************************************************************************/

void camera_setup(void)
{
	/* setup camera */
	camera_x=(WORLDX/2)<<SINSHIFT;
	camera_y=(WORLDY/2)<<SINSHIFT;
	camera_tilt=0;
}
void camera_move(unsigned char how)
{
unsigned short x,y;
	switch(how)
	{
		case 0: // forth
			x=camera_x+((SINUS2(camera_tilt))>>3);
			y=camera_y+((COSINUS2(camera_tilt))>>3);
			break;
		case 1: // back
			x=camera_x-((SINUS2(camera_tilt))>>3);
			y=camera_y-((COSINUS2(camera_tilt))>>3);
			break;
		case 2: // left
			break;
		case 3: // right
			break;
	}
	if(getblock(x,y)==0)
	{
		camera_x=x;
		camera_y=y;
#ifdef __C64__
		switch(how)
		{
			case 0: // forth
				irqincoffs();
				break;
			case 1: // back
				irqdecoffs();
				break;
		}
#endif
	}
}

/*************************************************************************************
 little benchmark
*************************************************************************************/

unsigned long fps;

void do_benchmark(void)
{
unsigned char i;
clock_t start,stop;

    start=clock();
	for(i=255;i!=0;--i)
	{
		// even frame
		cast_rays();
		draw_lines1();
		screen_setfb1();
		++camera_tilt;

        // odd frame
		cast_rays();
		draw_lines2();
		screen_setfb2();
        ++camera_tilt;
	}
	stop=clock();
	fps=(((512l<<8)*CLOCKS_PER_SEC)/(stop-start));
}

/*************************************************************************************
 main loop
*************************************************************************************/

unsigned char do_input(void)
{
	if(kbhit())
	{
		switch(cgetc())
		{
			case CH_CURS_UP:
				camera_move(0);
				break;
			case CH_CURS_DOWN:
				camera_move(1);
				break;
			case CH_CURS_LEFT:
				--camera_tilt;
				break;
			case CH_CURS_RIGHT:
				++camera_tilt;
				break;
			case CH_ENTER:
				return 0;
				break;
		}
	}
	return 1;
}

void do_freedir(void)
{
unsigned char frame=0;

	/* main loop */
	while(do_input())
	{
		if((frame^=1)==0)
		{
			// even frame
#ifdef ENABLESPRITES
	spriterender_start();
#endif
			cast_rays();
			draw_lines1();
			//draw_mirror1();
#ifdef ENABLESPRITES
	spriterender_finish();
#endif
			screen_setfb1();

			//draw_colmirror();
		}
		else
		{
			// odd frame
#ifdef ENABLESPRITES
	spriterender_start();
#endif

			cast_rays();
			draw_lines2();
			//draw_mirror2();
#ifdef ENABLESPRITES
	spriterender_finish();
#endif
			screen_setfb2();
			//draw_colmirror();
		}

	}

}

/*************************************************************************************
 menu
*************************************************************************************/

void do_menu(void)
{
	while(1)
	{
        bgcolor(COLOR_BLACK);
        clrscr();textcolor(COLOR_WHITE);cursor(1);revers(0);
		cputsxy(1,1,"minicast - very simple raycast demo");
		cputsxy(1,2,"(w)2004 groepaz/hitmen for cc65");
		cputsxy(1,4,"b - run benchmark");
		cputsxy(1,5,"f - free steering");
		cputsxy(1,6,"e - exit");
		cputsxy(1,12,"cursor keys to move (yes it sucks)");
		if(fps!=0)
		{
			gotoxy(1,10);
			cprintf("last fps: %ld.%ld",fps/256,((fps%256)/28));
		}
		cputsxy(1,8,"-=> ");
		switch(cgetc())
		{
			case 'b':
					screen_setup();
					camera_setup();
					do_benchmark();
					screen_reset();
				break;
			case 'f':
					screen_setup();
					camera_setup();
					do_freedir();
					screen_reset();
				break;
			case 'e':
					return;
				break;
		}
	}
}

/*************************************************************************************
 entry point
*************************************************************************************/

int main(void)
{
unsigned char oldbg,oldtx,oldbo;

    clrscr();
	oldbg=bgcolor(COLOR_BLACK);
	oldbo=bordercolor(COLOR_GRAY1);
    oldtx=textcolor(COLOR_WHITE);
    cursor(1);revers(0);
    cputsxy(1,1,"one moment please...");

    // call things that only need to be setup once
	fisheye_setup();
	world_setup();
	project_setup();
#ifdef __C64__
#ifdef ENABLESPRITES
	spriteinit();
#endif
#endif
	// call menu
	do_menu();

	bgcolor(oldbg);
	bordercolor(oldbo);
    textcolor(oldtx);
    cursor(0);revers(0);

	// exit like a good boy :)
	return EXIT_SUCCESS;
}
